/*
** Module   : SecurityOps.java
** Abstract : Progress 4GL compatible security interface
**
** Copyright (c) 2006-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 GES 20060331   @25287 First version implementing the USERID builtin
**                           function.
** 002 GES 20060331   @25294 Added ENCODE support (not Progress 4GL 
**                           compatible).
** 003 NVS 20060418   @25585 Added static convenience method to be called
**                           from the converted code: changePassword().
** 004 GES 20060831   @29063 Make sure encode output only includes chars
**                           from the set A-Z and a-z using a simple
**                           table driven base 52 approach.
** 005 GES 20101005          Added stub (non-functional) methods for the
**                           USERID(dbname) variant.
** 006 AIL 20121206          Added support for BASE64-ENCODE and 
**                           BASE64-DECODE  
*/

package com.goldencode.p2j.util;

import java.security.*;
import com.goldencode.p2j.security.SecurityManager;
import org.apache.commons.codec.binary.Base64;

/**
 * Provides static utility methods for Progress 4GL compatible security
 * functions.  At this time, the following Progress 4GL functions are
 * handled:
 * <p>
 * <pre>
 * Progress Function              Java Equivalent
 * ---------------------------    ---------------------------
 * USERID                         {@link #getUserId}
 * USERID(dbname)                 {@link #getUserIdFromDB}
 * ENCODE                         {@link #encode}
 * BASE64-ENCODE                  {@link #base64Encode}
 * BASE64-DECODE                  {@link #base64Decode} 
 * </pre>
 * <p>
 *
 * @author    GES
 */
public class SecurityOps
{   
   /** Table of all alphabetic characters used to limit the encode output. */
   private static final String ALPHAS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" +
                                        "abcdefghijklmnopqrstuvwxyz";
   
   /**
    * Returns the userid associated with the current user.
    * <p>
    * Note that in Progress 4GL, this takes an optional string parameter with
    * the name of the logical database in which the userid is set.  There is
    * no corresponding database level authentication concept in P2J so this
    * parameter is not supported.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserId() 
   {
      character userid = new character("");
      
      SecurityManager sm = SecurityManager.getInstance();
      
      if (sm != null)
      {
         userid = new character(sm.getUserId());
      }
      
      return userid;
   }
   
   /**
    * Returns the userid associated with the current user of the given logical
    * database.
    * <p>
    * This is just a stub at this time.  There is no corresponding database
    * level authentication concept in P2J so this is not YET working.
    *
    * @param    dbname
    *           Logical database name.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserIdFromDB(character dbname)
   {
      return getUserIdFromDB(dbname.getValue());
   }
   
   /**
    * Returns the userid associated with the current user of the given logical
    * database.
    * <p>
    * This is just a stub at this time.  There is no corresponding database
    * level authentication concept in P2J so this is not YET working.
    *
    * @param    dbname
    *           Logical database name.
    *
    * @return   The current user context's userid or the empty string in the
    *           case of any problem.
    */
   public static character getUserIdFromDB(String dbname)
   {
      return new character("");
   }
   
   /**
    * Performs a password change procedure for the currently logged on user.
    * <p>
    * Application has no control over the procedure itself. 
    *
    * @return   failure/success indication
    *           <code>true></code> means 
    */
   public static logical changePassword() 
   {
      logical result = new logical(false);

      SecurityManager sm = SecurityManager.getInstance();
      
      if (sm != null)
      {
         result = new logical(sm.changePassword());
      }
      
      return result;
   }

   /**
    * Hashes an arbitrary input string into a 16-byte output string.  This
    * conversion is one-way (the original data can not be recovered) and the
    * output is always a string of 16-bytes no matter the length of the input
    * string.
    * <p>
    * <b>WARNING: this implementation is NOT compatible with the Progress
    * 4GL algorithm in any way. Any programs relying on comparisons between
    * data encoded by the Progress 4GL ENCODE and the return value from this
    * method will NOT compare as expected.</b>
    *
    * @param    data
    *           The data to be hashed.  Must not be <code>null</code>.
    *
    * @return   The hashed 16-byte string or the <code>unknown value</code>
    *           if the input is <code>unknown value</code>.
    */
   public static character encode(character data)
   {
      if (data.isUnknown())
      {
         return new character();
      }
      
      return encode(data.getValue());
   }
   
   /**
    * Hashes an arbitrary input string into a 16-byte output string.  This
    * conversion is one-way (the original data can not be recovered) and the
    * output is always a string of 16-bytes no matter the length of the input
    * string.
    * <p>
    * <b>WARNING: this implementation is NOT compatible with the Progress
    * 4GL algorithm in any way. Any programs relying on comparisons between
    * data encoded by the Progress 4GL ENCODE and the return value from this
    * method will NOT compare as expected.</b>
    * <p>
    * The 16-bytes of output will only include characters in the range
    * <code>A-Z</code> and <code>a-z</code>.
    *
    * @param    data
    *           The data to be hashed.  Must not be <code>null</code>.
    *
    * @return   The hashed 16-byte string.
    */
   public static character encode(String data)
   {
      StringBuffer sb = new StringBuffer();
      byte[] hashed = null;
      
      try
      {
         MessageDigest md = MessageDigest.getInstance("MD5");
         hashed = md.digest(data.getBytes());
         
         for (int i = 0; i < hashed.length; i++)
         {
            // this is not a fully even distribution but it is close
            // enough for an interrim solution
            int base52 = ((0x000000FF & hashed[i]) % 52); 
            sb.append(ALPHAS.charAt(base52));
         }
      }
      
      catch (NoSuchAlgorithmException nsa)
      {
         // we will fail below
      }
      
      if (hashed == null || hashed.length != 16)
      {
         throw new RuntimeException("Unexpected hashing result for input '" +
                                    data + "' resulting in output '" +
                                    hashed + "'.");
      }
      
      return new character(sb.toString());
   }   
   
   /**
    * Converts binary data into a Base64 encoded string and returns a 
    * longchar with the encoded string.
    *
    * @param    data
    *           The data to be encoded.
    *
    * @return   a longchar with the encoded string.
    */
   public static longchar base64Encode(BinaryData data)
   {
      if (data.isUnknown())
      {
         return new longchar();
      }

      return base64Encode(data.getValue());
   }

   /**
   * Converts binary data into a Base64 encoded string and returns a 
   * longchar with the encoded string.
   *
   * @param    data
   *           The data to be encoded.
   *
   * @return   a longchar with the encoded string.
   */
   public static longchar base64Encode(byte[] data)
   {
      byte[] encoded = Base64.encodeBase64(data);

      return new longchar(new String(encoded));
   }

   /**
    * Converts a base64 encoded string into a binary value and returns a 
    * memptr with the decoded value.
    *
    * @param    data
    *           The string to be decoded.
    *
    * @return   a memptr with the decoded data.
    */
   public static memptr base64Decode(longchar data)
   {
      //TODO - review
      if (data.isUnknown()) 
      {
         return new memptr();
      }

      return base64Decode(data.getValue());
   }

   /**
    * Converts a base64 encoded string into a binary value and returns a 
    * memptr with the decoded value.
    *
    * @param    data
    *           The string to be decoded.
    *
    * @return   a memptr with the decoded data.
    */
   public static memptr base64Decode(String data)
   {
      byte[] decoded = Base64.decodeBase64(data.getBytes());
      memptr m = new memptr();

      m.setValue(decoded);

      return m;
   }
}
