/*
** Module   : TextOps.java
** Abstract : Progress 4GL compatible TextOps object
**
** Copyright (c) 2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------Description----------------------
** 001 AIL 20121206 Created initial version by moving static methods from 
**                  the character class as this methods will be used with 
**                  character and also longchar arguments.
*/

package com.goldencode.p2j.util;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A class that provides Progress 4GL compatible character and longchar 
 * functions and operators.  This class provides all functionality as statics
 * and has no instance data.
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * num-entries      function   {@link #numEntries}
 * begins           operator   {@link #begins} and {@link #_begins}
 * matches          operator   {@link #matches} and {@link #_matches} 
 * can-do           function   {@link #matchesList}
 * caps             function   {@link #toUpperCase}
 * compare          function   {@link #compare}
 * entry            function   {@link #entry}
 * index            function   {@link #indexOf}
 * lc               function   {@link #toLowerCase}
 * left-trim        function   {@link #leftTrim}
 * length           function   {@link #length} and {@link #byteLength}
 * lookup           function   {@link #lookup}
 * r-index          function   {@link #lastIndexOf}
 * replace          function   {@link #replaceAll}
 * right-trim       function   {@link #rightTrim}
 * substitute       function   {@link #substitute}
 * substring        function   {@link #substring}
 * trim             function   {@link #trim}
 * </pre>
 * <p>
 * There are convenience versions of the begins and matches operators that
 * directly return <code>boolean</code> values such that unwrapping is not
 * needed.  These are named the same as the operator methods above except 
 * there is no 'With' appended.  For example, the begins operator is
 * {@link #begins}.
 * <p>
 * This class implements unknown value logic for all operators and functions.
 * Generally, if any parameter of the these methods is unknown, the result is
 * unknown.
 * <p>
 *
 * @author    AIL
 */
public final class TextOps
{

   /** Specifies the less than operator for {@link #compare}. */
   private static final int LT = 0;
   
   /** Specifies the less than or equal operator for {@link #compare}. */
   private static final int LE = 1;
   
   /** Specifies the greater than operator for {@link #compare}. */
   private static final int GT = 2;
   
   /** Specifies the greater than or equal operator {@link #compare}. */
   private static final int GE = 3;
   
   /** Specifies the equal operator for {@link #compare}. */
   static final int EQ = 4;
   
   /** Specifies the not equal operator for {@link #compare}. */
   private static final int NE = 5;
   
   /** Specifies the RAW type for {@link #compare} or {@link substring}. */
   private static final int RAW = 6;
   
   /** 
    * Specifies the COLUMN type for {@link #compare} or {@link substring}.
    */
   private static final int COLUMN = 7;
   
   /** Specifies the FIXED type for {@link #compare} or {@link substring}. */
   private static final int FIXED = 8;
   
   /**
    * Specifies the CHARACTER type for {@link #compare} or {@link substring}.
    */      
   private static final int CHARACTER = 9;
   
   /** Specifies the CASE-SENSITIVE strength for {@link #compare}. */
   static final int CASE_SENSITIVE = 10;
   
   /** Specifies the CASE-INSENSITIVE strength for {@link #compare}. */
   private static final int CASE_INSENSITIVE = 11;

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, basing case sensitivity on the left operand.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    *
    * @return   <code>true</code> if op matches the regular expression or
    *           <code>false</code> if there is no match or if either
    *           operand is <code>unknown</code>.
    */
   public static boolean _matches(character op, character pattern)
   {
      if (op.isUnknown() || pattern.isUnknown())
         return false;
      
      return _matches(op.getValue(), pattern.getValue(), op.isCaseSensitive());
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, basing case sensitivity on the left operand.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    *
    * @return   <code>true</code> if op matches the regular expression or
    *           <code>false</code> if there is no match or if the left
    *           operand is <code>unknown</code>.
    */
   public static boolean _matches(character op, String pattern)
   {
      if (op.isUnknown())
         return false;
      
      return _matches(op.getValue(), pattern, op.isCaseSensitive());
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, case-insensitively.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    *
    * @return   <code>true</code> if op matches the regular expression or
    *           <code>false</code> if there is no match or if the right
    *           operand is <code>unknown</code>.
    */
   public static boolean _matches(String op, character pattern)
   {
      if (pattern.isUnknown())
         return false;
      
      return _matches(op, pattern.getValue(), false);
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, ignoring the case of the operand and the
    * pattern.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    *
    * @return   <code>true</code> if op matches the regular expression.
    */
   public static boolean _matches(String op, String pattern)
   {
      return _matches(op, pattern, false);
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, honoring the case sensitivity preference of
    * the caller.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    * @param    caseSens
    *           <code>true</code> if a case-sensitive comparison should be
    *           used.
    *
    * @return   <code>true</code> if op matches the regular expression.
    */
   public static boolean _matches(String op, String pattern, boolean caseSens)
   {
      String up = caseSens ? op : op.toUpperCase();
      
      pattern = caseSens ? pattern : pattern.toUpperCase();
      return up.matches(character.convertToRegEx(pattern));
   }

   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if either of the two operands is
    * case-sensitive.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2 or 
    *           <code>false</code> if there is no match or if either
    *           operand is <code>unknown value</code>.
    */
   public static boolean _begins(character op1, character op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return false;
      
      boolean cs = (op1.isCaseSensitive() || op2.isCaseSensitive());
      
      return _begins(op1.getValue(), op2.getValue(), cs);
   }

   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if the right operand is case-sensitive.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2 or 
    *           <code>false</code> if there is no match or if the right
    *           operand is <code>unknown value</code>.
    */
   public static boolean _begins(String op1, character op2)
   {
      if (op2.isUnknown())
         return false;
      
      return _begins(op1, op2.getValue(), op2.isCaseSensitive());
   }

   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if the left operand is case-sensitive.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2 or 
    *           <code>false</code> if there is no match or if the left
    *           operand is <code>unknown value</code>.
    */
   public static boolean _begins(character op1, String op2)
   {
      if (op1.isUnknown())
         return false;
      
      return _begins(op1.getValue(), op2, op1.isCaseSensitive());
   }

   /**
    * Tests if the first operand begins with the second operand, ignoring the
    * case of the two operands.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2.
    */
   public static boolean _begins(String op1, String op2)
   {
      return _begins(op1, op2, false);
   }

   /**
    * Tests if the first operand begins with the second operand, honoring the
    * case sensitivity preference of the caller.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    * @param    caseSens
    *           <code>true</code> if a case-sensitive comparison should be
    *           used.
    *
    * @return   <code>true</code> if op1 begins with op2.
    */
   public static boolean _begins(String op1, String op2, boolean caseSens)
   {
      String up1 = caseSens ? op1 : op1.toUpperCase();
      String up2 = caseSens ? op2 : op2.toUpperCase();
      
      return up1.startsWith(up2);
   }

   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if either of the two operands is
    * case-sensitive.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2 or 
    *           <code>unknown value</code> if either operand is
    *           <code>unknown value</code>.
    */
   public static logical begins(character op1, character op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return new logical();
      
      boolean cs = (op1.isCaseSensitive() || op2.isCaseSensitive());
      
      return begins(op1.getValue(), op2.getValue(), cs);
   }

   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if the right operand is case-sensitive.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2 or 
    *           <code>unknown value</code> if the right operand is
    *           <code>unknown value</code>.
    */
   public static logical begins(String op1, character op2)
   {
      if (op2.isUnknown())
         return new logical();
      
      return begins(op1, op2.getValue(), op2.isCaseSensitive());
   }

   /**
    * Tests if the first operand begins with the second operand, implementing
    * a case-sensitive comparison if the left operand is case-sensitive.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2 or 
    *           <code>unknown value</code> if the left operand is
    *           <code>unknown value</code>.
    */
   public static logical begins(character op1, String op2)
   {
      if (op1.isUnknown())
         return new logical();
      
      return begins(op1.getValue(), op2, op1.isCaseSensitive());
   }

   /**
    * Tests if the first operand begins with the second operand, ignoring the
    * case of the two operands.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    *
    * @return   <code>true</code> if op1 begins with op2.
    */
   public static logical begins(String op1, String op2)
   {
      return begins(op1, op2, false);
   }

   /**
    * Tests if the first operand begins with the second operand, honoring the
    * case sensitivity preference of the caller.
    *
    * @param    op1
    *           First string to compare.
    * @param    op2
    *           Second string to compare.
    * @param    caseSens
    *           <code>true</code> if a case-sensitive comparison should be
    *           used.
    *
    * @return   <code>true</code> if op1 begins with op2.
    */
   public static logical begins(String op1, String op2, boolean caseSens)
   {
      String up1 = caseSens ? op1 : op1.toUpperCase();
      String up2 = caseSens ? op2 : op2.toUpperCase();
      
      return new logical(up1.startsWith(up2));
   }

   /**
    * Compares two character values with a runtime-specified strength and
    * operator.  This method will probably generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for conversion.
    * The result should match up with the Progress equivalents in the ASCII
    * range.
    * <p>
    * This method does not handle the case of user-specified collation tables
    * as is possible in Progress.
    * <p>
    * In addition, the case-sensitive and case-insensitive strength options
    * are likely to differ in some results since the current implementation
    * really just implements a direct numeric comparison based on each 
    * UNICODE character's value (though case-insensitive support is
    * provided but it just converts everything to the same case before
    * implementing the 'raw' lexicographic comparison.
    *
    * @param    op1
    *           The left operand.
    * @param    operator
    *           One of {@link character#LT}, {@link character#LE},
    *           {@link character#GT}, {@link character#GE},
    *           {@link character#EQ} or {@link character#NE}.
    * @param    op2
    *           The right operand.
    * @param    strength
    *           One of {@link character#RAW}, {@link character#CASE_SENSITIVE}
    *           or {@link character#CASE_INSENSITIVE}.
    *
    * @return   The result of the condition evaluation. 
    */   
   public static logical compare(character op1,
                                 int       operator,
                                 character op2,
                                 int       strength)
   {
      if (operator != LT && operator != LE && operator != GT &&
          operator != GE && operator != EQ && operator != NE) 
      {
         // unrecognized operator
         return new logical();
      }
      
      // unknown value logic
      if (op1.isUnknown() && op2.isUnknown())
      {
         // both are unknown
         if (operator == LE || operator == GE || operator == EQ) 
         {
            return new logical(true);
         }
         else
         {
            return new logical(false);
         }
      }
      else
      {
         // both are NOT unknown BUT...
         if (op1.isUnknown() || op2.isUnknown())
         {
            // one or the other is unknown
            if (operator == NE)
            {
               return new logical(true);
            }
            else
            {
               return new logical(false);
            }
         }
      }
      
      int result;
      
      if (strength == RAW || strength == CASE_SENSITIVE)
      {
         // RAW and CASE_SENSITIVE are the same right now
         result = op1.getValue().compareTo(op2.getValue());
      }
      else if (strength == CASE_INSENSITIVE) 
      {
         result = op1.getValue().compareToIgnoreCase(op2.getValue());
      }
      else
      {
         // unrecognized strength
         return new logical();
      }
      
      boolean conclude = false;
      
      switch (operator)
      {
         case LT:
            conclude = (result < 0);
            break;
         case LE:
            conclude = (result <= 0);
            break;
         case GT:
            conclude = (result > 0);
            break;
         case GE:
            conclude = (result >= 0);
            break;
         case EQ:
            conclude = (result == 0);
            break;
         case NE:
            conclude = (result != 0);
            break;
      }
      
      return new logical(conclude);
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i,
                                 character  list,
                                 character  delimit)
   throws ErrorConditionException
   {
      if (i.isUnknown() || list.isUnknown() || delimit.isUnknown())
      {
         return new character();
      }
      
      return entry(i.intValue(), list.getValue(), TextOps.safeDelimiter(delimit.getValue()));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double i, character list, character delimit)
   throws ErrorConditionException
   {
      if (list.isUnknown() || delimit.isUnknown())
      {
         return new character();
      }
      
      return entry((int) i, list.getValue(), TextOps.safeDelimiter(delimit.getValue()));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i, String list, character delimit)
   throws ErrorConditionException
   {
      if (i.isUnknown() || delimit.isUnknown())
      {
         return new character();
      }
      
      return entry(i.intValue(), list, TextOps.safeDelimiter(delimit.getValue()));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i,
                                 character  list,
                                 String     delimit)
   throws ErrorConditionException
   {
      if (i.isUnknown() || list.isUnknown())
      {
         return new character();
      }
      
      return entry(i.intValue(), list.getValue(), TextOps.safeDelimiter(delimit));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double i, character list, String delimit)
   throws ErrorConditionException
   {
      if (list.isUnknown())
      {
         return new character();
      }
      
      return entry((int) i, list.getValue(), TextOps.safeDelimiter(delimit));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i, String list, String delimit)
   throws ErrorConditionException
   {
      if (i.isUnknown())
      {
         return new character();
      }
      
      return entry(i.intValue(), list, TextOps.safeDelimiter(delimit));
   }

   /**
    * Returns the specified 1-based indexed comma-delimited entry in the
    * passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i, character list)
   throws ErrorConditionException
   {
      if (i.isUnknown() || list.isUnknown())
      {
         return new character();
      }
      
      return entry(i.intValue(), list.getValue(), ',');
   }

   /**
    * Returns the specified 1-based indexed comma-delimited entry in the
    * passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(NumberType i, String list)
   throws ErrorConditionException
   {
      if (i.isUnknown())
      {
         return new character();
      }
      
      return entry(i.intValue(), list, ',');
   }

   /**
    * Returns the specified 1-based indexed comma-delimited entry in the
    * passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double i, character list)
   throws ErrorConditionException
   {
      if (list.isUnknown())
      {
         return new character();
      }
      
      return entry((int) i, list.getValue(), ',');
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double i, String list, String delimit)
   throws ErrorConditionException
   {
      return entry((int) i, list, TextOps.safeDelimiter(delimit));
   }

   /**
    * Returns the specified 1-based indexed entry in the passed list.  
    *
    * @param    ii
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double ii, String list, char delimit)
   throws ErrorConditionException
   {
      // create the regular expression (with proper escapes)
      String regex = fixupRegexChar(delimit);      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = list.split(regex, -1);
      
      // convert 1 based index to 0 based
      int i   = (int) ii;
      int idx = i - 1;
      
      if (idx < 0 || idx >= array.length)
      {
         genOutsideRangeError(i, list);
         
         // this is the safest thing to return (it should be undone anyway)
         return new character("");
      }
      
      return new character(array[idx]);
   }

   /**
    * Returns the specified 1-based indexed comma-delimited entry in the
    * passed list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The entry in the list or the <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static character entry(double i, String list)
   throws ErrorConditionException
   {
      return entry((int) i, list, ',');
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    c
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(character c, NumberType num)
   {
      if (c.isUnknown() || num.isUnknown())
      {
         return new character();
      }
      
      return fill(c.getValue(), num.intValue());
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    c
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(character c, int num)
   {
      if (c.isUnknown())
      {
         return new character();
      }
      
      return fill(c.getValue(), num);
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    c
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(character c, double num)
   {
      if (c.isUnknown())
      {
         return new character();
      }
      
      return fill(c.getValue(), (int) num);
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    str
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(String str, NumberType num)
   {
      if (num.isUnknown())
      {
         return new character();
      }
      
      return fill(str, num.intValue());
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    str
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(String str, double num)
   {
      return fill(str, (int) num);
   }

   /**
    * Return a character instance that is comprised of <code>num</code>
    * concatinated copies of a given string. 
    *
    * @param    str
    *           The string to be repeated.
    * @param    num
    *           The number of times to repeat.
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code> or an empty string if 
    *           <code>num</code> < 1.
    */
   public static character fill(String str, int num)
   {
      if (num < 1)
         return new character("");
      
      StringBuffer sb = new StringBuffer();
      
      for (int i = 0; i < num; i++)
      {
         sb.append(str);
      }
      
      return new character(sb.toString());
   }

   /**
    * Returns the length of the given <code>character</code> instance in
    * bytes.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of bytes in the given instance or -1 if the
    *           instance is <code>unknown value</code>.
    */
   public static int byteLengthOf(character c)
   {
      if (c.isUnknown())
         return -1;
      
      return c.getValue().getBytes().length;
   }

   /**
    * Returns the length of the given <code>character</code> instance in
    * bytes.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of bytes in the given instance or
    *           <code>unknown value</code> if the instance is
    *           <code>unknown value</code>.
    */
   public static integer byteLength(character c)
   {
      if (c.isUnknown())
         return new integer();
      
      return new integer(c.getValue().getBytes().length);
   }

   /**
    * Generate an error for a substring or overlay index that is negative.
    */
   static void genNegativeIndexError()
   {
      String err = "Starting position for SUBSTRING, OVERLAY, etc. must" +
                   " be 1 or greater";
      ErrorManager.recordOrThrowError(82, err);
   }

   /**
    * Generate an error for a value out of range.
    *
    * @param    idx
    *           The index that is out of range.
    * @param    list
    *           The list being indexed.
    */
   static void genOutsideRangeError(int idx, String list)
   {
      String spec = "Entry %d is outside the range of list %s";
      String err  = String.format(spec, idx, list);
      
      ErrorManager.recordOrThrowError(560, err);
   }

   /**
    * Converts a single character into a regular expression, including
    * escaping any of the special characters that have other meaning in
    * regular expressions.
    *
    * @param    ch
    *           The exact character on which to match.
    *
    * @return   The resulting regular expression.
    */
   static String fixupRegexChar(char ch)
   {
      StringBuffer regex = new StringBuffer();
      
      if (ch == '(' || ch == ')' || ch == '.' || ch == '*' || ch == '[' ||
          ch == ']' || ch == '^' || ch == '{' || ch == '}' || ch == '?' ||
          ch == '+' || ch == '?' || ch == '$' || ch == '|' || ch == '\\')
      {
         regex.append('\\').append(ch);
      }
      else
      {
         regex.append(ch);
      }
      
      return regex.toString();
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(character  src,
                                 character  target,
                                 NumberType i)
   {
      if (src.isUnknown() || target.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(),
                     target.getValue(),
                     i.doubleValue(),
                     src.isCaseSensitive() || target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(String     src,
                                 character  target,
                                 NumberType i)
   {
      if (target.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src,
                     target.getValue(),
                     i.doubleValue(),
                     target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(character  src,
                                 String     target,
                                 NumberType i)
   {
      if (src.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(),
                     target,
                     i.doubleValue(),
                     src.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(String     src,
                                 String     target,
                                 NumberType i)
   {
      if (i.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src, target, i.doubleValue(), false);
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(character  src,
                                 character  target,
                                 double     i)
   {
      if (src.isUnknown() || target.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(),
                     target.getValue(),
                     i,
                     src.isCaseSensitive() || target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(String     src,
                                 character  target,
                                 double     i)
   {
      if (target.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src, target.getValue(), i, target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(character  src,
                                 String     target,
                                 double     i)
   {
      if (src.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(), target, i, src.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(character src,
                                 character target)
   {
      if (src.isUnknown() || target.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(),
                     target.getValue(),
                     1,
                     src.isCaseSensitive() || target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(String    src,
                                 character target)
   {
      if (target.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src, target.getValue(), 1, target.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer indexOf(character src,
                                 String    target)
   {
      if (src.isUnknown())
      {
         return new integer();
      }
      
      return indexOf(src.getValue(), target, 1, src.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    ii
    *           The (1-based) starting index from which to search.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer indexOf(String src, String target, double ii)
   {
      return indexOf(src, target, ii, false);
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer indexOf(String src, String target)
   {
      return indexOf(src, target, 1, false);
   }

   /**
    * Returns the 1-based index of one character expression within another
    * given character expression based on a known starting point for the
    * search.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    ii
    *           The (1-based) starting index from which to search.
    * @param    caseSensitive
    *           <code>true</code> if case must be considered in the search;
    *           <code>false</code> if the search is case insensitive.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer indexOf(String src,
                                 String target,
                                 double ii,
                                 boolean caseSensitive)
   {
      int i = (int) ii;
      
      if (src.length() == 0 || target.length() == 0 || i < 1)
      {
         return new integer(0);
      }
      
      if (!caseSensitive)
      {
         src = src.toLowerCase();
         target = target.toLowerCase();
      }
      
      return new integer(src.indexOf(target, i - 1) + 1);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(character  src,
                                     character  target,
                                     NumberType i)
   {
      if (src.isUnknown() || target.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target.getValue(), i.intValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(String     src,
                                     character  target,
                                     NumberType i)
   {
      if (target.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src, target.getValue(), i.intValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(character  src,
                                     String     target,
                                     NumberType i)
   {
      if (src.isUnknown() || i.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target, i.intValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(String     src,
                                     String     target,
                                     NumberType i)
   {
      if (i.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src, target, i.intValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(character src,
                                     character target,
                                     double    i)
   {
      if (src.isUnknown() || target.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target.getValue(), (int) i);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(String    src,
                                     character target,
                                     double    i)
   {
      if (target.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src, target.getValue(), (int) i);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    i
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(character src,
                                     String    target,
                                     double    i)
   {
      if (src.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target, (int) i);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning).
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(character src, character target)
   {
      if (src.isUnknown() || target.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target.getValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning).
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(String src, character target)
   {
      if (target.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src, target.getValue());
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning).
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found or <code>unknown value</code>
    *           if any input is <code>unknown value</code>.
    */
   public static integer lastIndexOf(character src, String target)
   {
      if (src.isUnknown())
      {
         return new integer();
      }
      
      return lastIndexOf(src.getValue(), target);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning) starting at a specified
    * starting index.
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    * @param    ii
    *           The starting index to search from.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer lastIndexOf(String src, String target, double ii)
   {
      int i = (int) ii;
      
      if (src.length() == 0 || target.length() == 0 || i < 1)
      {
         return new integer(0);
      }
      
      return new integer(src.lastIndexOf(target, i - 1) + 1);
   }

   /**
    * Returns the last 1-based index of one character expression within
    * another given character expression (the search starts at the end
    * of the string and moves toward the beginning).
    *
    * @param    src
    *           The character expression to search in.
    * @param    target
    *           The character expression to search for.
    *
    * @return   The 1-based index in the source at which the target is found
    *           or 0 if the target is not found.
    */
   public static integer lastIndexOf(String src, String target)
   {
      return lastIndexOf(src, target, src.length() + 1);
   }

   /**
    * Returns the length of the given <code>character</code> instance in
    * characters.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of characters in the given instance or -1 if the
    *           instance is <code>unknown value</code>.
    */
   public static int lengthOf(character c)
   {
      if (c.isUnknown())
         return -1;
      
      return c.getValue().length();
   }

   /**
    * Returns the length of the given <code>character</code> instance in
    * characters.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of characters in the given instance
    *           <code>unknown value</code> if the instance is
    *           <code>unknown value</code>.
    */
   public static integer length(character c)
   {
      if (c.isUnknown())
         return new integer();
      
      return new integer(c.getValue().length());
   }

   /**
    * Returns the length of the given string instance in characters.
    *
    * @param    c
    *           The instance to test.
    *
    * @return   The number of characters in the given instance.
    */
   public static integer length(String c)
   {
      return new integer(c.length());
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of only
    * the input parameter, the trim character list case-sensitivity is
    * ignored.  Likewise, if the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(character input, character list)
   {
      if (input.isUnknown() || list.isUnknown())
      {
         return new character();
      }
      
      return leftTrim(input.getValue(), list.getValue(), input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * the input parameter. If the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(character input, String list)
   {
      if (input.isUnknown())
      {
         return new character();
      }
      
      return leftTrim(input.getValue(), list, input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.
    * <p>
    * Trim operations are performed case-insensitively.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(String input, character list)
   {
      if (list.isUnknown())
      {
         return new character();
      }
      
      return leftTrim(input, list.getValue(), false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * left of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * input parameter.  Likewise, if the input parameter is case-sensitive,
    * then the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(character input)
   {
      if (input.isUnknown())
      {
         return new character();
      }
      
      return leftTrim(input.getValue(), null, input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(String  input,
                                    String  list,
                                    boolean caseSens)
   {
      return new character(TextOps.trim(input, list, true, caseSens), caseSens);
   }

   /**
    * Trims the list of contiguous trim characters from the left of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(String input, String list)
   {
      return leftTrim(input, list, false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * left of an input string.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character leftTrim(String input)
   {
      return leftTrim(input, (String) null);
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, basing case sensitivity on the left operand.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    *
    * @return   <code>true</code> if op matches the regular expression or
    *           <code>unknown value</code> if either operand is
    *           <code>unknown</code>.
    */
   public static logical matches(character op, character pattern)
   {
      if (op.isUnknown() || pattern.isUnknown())
         return new logical();
      
      return matches(op.getValue(), pattern.getValue(), op.isCaseSensitive());
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, basing case sensitivity on the left operand.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    *
    * @return   <code>true</code> if op matches the regular expression or
    *           <code>unknown value</code> if the left operand is
    *           <code>unknown</code>.
    */
   public static logical matches(character op, String pattern)
   {
      if (op.isUnknown())
         return new logical();
      
      return matches(op.getValue(), pattern, op.isCaseSensitive());
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, case-insensitively.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    *
    * @return   <code>true</code> if op matches the regular expression or
    *           <code>unknown value</code> if the right operand is
    *           <code>unknown</code>.
    */
   public static logical matches(String op, character pattern)
   {
      if (pattern.isUnknown())
         return new logical();
      
      return matches(op, pattern.getValue(), false);
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, ignoring the case of the operand and the
    * pattern.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    *
    * @return   <code>true</code> if op matches the regular expression.
    */
   public static logical matches(String op, String pattern)
   {
      return matches(op, pattern, false);
   }

   /**
    * Tests if the first operand matches a regular expression that is 
    * provided by the caller, honoring the case sensitivity preference of
    * the caller.
    * <p>
    * The Progress style matching pattern passed as a parameter is converted
    * to a regular expression using {@link character#convertToRegEx} during this
    * method.
    *
    * @param    op
    *           String to check for a match.
    * @param    pattern
    *           Progress style matching pattern.
    * @param    caseSens
    *           <code>true</code> if a case-sensitive comparison should be
    *           used.
    *
    * @return   <code>true</code> if op matches the regular expression.
    */
   public static logical matches(String  op,
                                 String  pattern,
                                 boolean caseSens)
   {
      String up = caseSens ? op : op.toUpperCase();
      
      pattern = caseSens ? pattern : pattern.toUpperCase();
      boolean result = up.matches(character.convertToRegEx(pattern));
      
      return new logical(result);
   }

   /**
    * Tests if the current PROGRESS user is allowed in the list.
    * Since in the P2J environment the security subsystem is based on the
    * Security Manager, this call simply catches the overlooked CAN-DO call
    * in the PROGRESS code.
    *
    * @param    list
    *           comma-separated list of user IDs, possibly with ! and *
    *           modifiers.
    *
    * @return   never
    * @throws   IllegalArgumentException
    *           to catch the untranslated CAN-DO call
    */
   public static logical matchesList(character list)
   throws IllegalArgumentException
   {
      throw new IllegalArgumentException("untranslated CAN-DO call");
   }

   /**
    * Tests if the current PROGRESS user is allowed in the list.
    * Since in the P2J environment the security subsystem is based on the
    * Security Manager, this call simply catches the overlooked CAN-DO call
    * in the PROGRESS code.
    *
    * @param    list
    *           comma-separated list of user IDs, possibly with ! and *
    *           modifiers.
    *
    * @return   never
    * @throws   IllegalArgumentException
    *           to catch the untranslated CAN-DO call
    */
   public static logical matchesList(String list)
   throws IllegalArgumentException
   {
      throw new IllegalArgumentException("untranslated CAN-DO call");
   }

   /**
    * Tests if a string specified with the "item" parameter is allowed in the 
    * comma-separated list of allowed items.
    *
    * @param    list
    *           comma-separated list of allowed items, possibly with ! and *
    *           modifiers.
    * @param    item
    *           an item being tested
    *
    * @return   <code>true</code> if item is allowed in the list
    */
   public static logical matchesList(character list, character item)
   {
      if (list.isUnknown() || item.isUnknown())
         return new logical();
   
      return matchesList(list.getValue(), item.getValue());
   }

   /**
    * Tests if a string specified with the "item" parameter is allowed in the 
    * comma-separated list of allowed items.
    *
    * @param    list
    *           comma-separated list of allowed items, possibly with ! and *
    *           modifiers.
    * @param    item
    *           an item being tested
    *
    * @return   <code>true</code> if item is allowed in the list
    */
   public static logical matchesList(character list, String item)
   {
      if (list.isUnknown())
         return new logical();
   
      return matchesList(list.getValue(), item);
   }

   /**
    * Tests if a string specified with the "item" parameter is allowed in the 
    * comma-separated list of allowed items.
    *
    * @param    list
    *           comma-separated list of allowed items, possibly with ! and *
    *           modifiers.
    * @param    item
    *           an item being tested
    *
    * @return   <code>true</code> if item is allowed in the list
    */
   public static logical matchesList(String list, character item)
   {
      if (item.isUnknown())
         return new logical();
   
      return matchesList(list, item.getValue());
   }

   /**
    * Tests if a string specified with the "item" parameter is allowed in the 
    * comma-separated list of allowed items.
    *
    * @param    list
    *           comma-separated list of allowed items, possibly with ! and *
    *           modifiers.
    * @param    item
    *           an item being tested
    *
    * @return   <code>true</code> if item is allowed in the list
    */
   public static logical matchesList(String list, String item)
   {
      boolean result = false;
   
      // create the regular expression (with proper escapes)
      String regex = fixupRegexChar(',');      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = list.split(regex, -1);
      
      // checking item against every entry in the list until match is found
      int idx = 0;
      int i = 0;
      boolean negation = false;
      String piece = null;
      StringBuffer sb = new StringBuffer();
      char c = 0;
      Pattern p = null;
      Matcher m = null;
   
      for (idx = 0; idx < array.length; idx ++)
      {
         // isolate negation (!) case
         if (array[idx].length() > 0 && array[idx].charAt(0) == '!')
         {
            piece = array[idx].substring(1);
            negation = true;
         }
         else
         {
            piece = array[idx];
            negation = false;
         }
   
         // build a regex out of the list entry; only * is converted into .*
         sb.setLength(0);
         for (i = 0; i < piece.length(); i ++)
         {
            c = piece.charAt(i);
            if (c != '*')
               sb.append(fixupRegexChar(c));
            else
               sb.append(".*");
         }
         regex = sb.toString();
   
         // compile regex as case-insensitive and apply it to the item
         p = Pattern.compile(regex, Pattern.CASE_INSENSITIVE);
         m = p.matcher(item);
         if (!m.matches())
            continue;
   
         // this is a matching entry from the list; produce the result
         result = !negation;
         break;
      }
   
      return new logical(result);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, character list, character delim)
   {
      if (c.isUnknown() || list.isUnknown() || delim.isUnknown())
      {
         return new integer();
      }
      
      boolean cs = (c.isCaseSensitive() || list.isCaseSensitive());
      
      return lookup(c.getValue(), list.getValue(), TextOps.safeDelimiter(delim.getValue()), cs);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, String list, character delim)
   {
      if (c.isUnknown() || delim.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c.getValue(), list, TextOps.safeDelimiter(delim.getValue()), c.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, character list, character delim)
   {
      if (list.isUnknown() || delim.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c, list.getValue(), TextOps.safeDelimiter(delim.getValue()), list.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, character list, String delim)
   {
      if (c.isUnknown() || list.isUnknown())
      {
         return new integer();
      }
      
      boolean cs = (c.isCaseSensitive() || list.isCaseSensitive());
      
      return lookup(c.getValue(), list.getValue(), TextOps.safeDelimiter(delim), cs);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * comma-delimited list.  The search will be case-sensitive if either
    * argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, character list)
   {
      if (c.isUnknown() || list.isUnknown())
      {
         return new integer();
      }
      
      boolean cs = (c.isCaseSensitive() || list.isCaseSensitive());
      
      return lookup(c.getValue(), list.getValue(), ',', cs);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * comma-delimited list.  The search will be case-sensitive if either
    * argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, String list)
   {
      if (c.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c.getValue(), list, ',', c.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * comma-delimited list.  The search will be case-sensitive if either
    * argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, character list)
   {
      if (list.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c, list.getValue(), ',', list.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter.  The search will be done
    * case-insensitively.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found.
    */
   public static integer lookup(String c, String list, char delim)
   {
      return lookup(c, list, delim, false);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter.  
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    * @param    cs
    *           <code>true</code> specifies the match is to be done with
    *           case-sensitivity.
    *
    * @return   The 1-based index into the list or 0 if no match is found.
    */
   public static integer lookup(String c, String list, char delim, boolean cs)
   {
      // create the regular expression (with proper escapes)
      String regex = fixupRegexChar(delim);      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = list.split(regex, -1);
      
      // only use up to the first instance of the delimiter in the search 
      // pattern
      int    idx    = c.indexOf(delim);
      String search = idx < 0 ? c : c.substring(0, idx); 
      
      // default return value
      int index  = 0;
      
      // do our search
      for (int i = 0; i < array.length; i++)
      {
         if (cs ? array[i].equals(search) : array[i].equalsIgnoreCase(search))
         {
            // match found!  convert from 0-based to 1-based
            index = i + 1;
            break;
         }
      }
      
      return new integer(index);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * comma-delimited list.  The search will be done case-insensitively.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, String list)
   {
      return lookup(c, list, ',', false);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, String list, String delim)
   {
      return lookup(c, list, TextOps.safeDelimiter(delim), false);
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, character list, String delim)
   {
      if (list.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c, list.getValue(), TextOps.safeDelimiter(delim), list.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(character c, String list, String delim)
   {
      if (c.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c.getValue(), list, TextOps.safeDelimiter(delim), c.isCaseSensitive());
   }

   /**
    * Returns the 1-based index of a specified entry in the passed 
    * list with a user-specified delimiter. The search will be case-sensitive
    * if either argument is case-sensitive.
    *
    * @param    c
    *           The entry to find in the list.
    * @param    list
    *           The delimited list of entries.
    * @param    delim
    *           The delimiter character used to parse the list into entries.
    *
    * @return   The 1-based index into the list or 0 if no match is found or
    *           the <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public static integer lookup(String c, String list, character delim)
   {
      if (delim.isUnknown())
      {
         return new integer();
      }
      
      return lookup(c, list, TextOps.safeDelimiter(delim.getValue()), false);
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or -1 if the list is the
    *           <code>unknown value</code>.
    */
   public static int numEntriesOf(character list, character delimit)
   {
      if (list.isUnknown() || delimit.isUnknown())
      {
         return -1;
      }
      
      return numEntriesOf(list.getValue(), TextOps.safeDelimiter(delimit.getValue()));
   }

   /**
    * Returns the number of comma-delimited entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The number of entries in the list or -1 if the list is the
    *           <code>unknown value</code>.
    */
   public static int numEntriesOf(character list)
   {
      if (list.isUnknown())
      {
         return -1;
      }
      
      return numEntriesOf(list.getValue(), ',');
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list.
    */
   public static int numEntriesOf(String list, char delimit)
   {
      if (list.length() == 0)
      {
         return 0;
      }
      
      // create the regular expression (with proper escapes)
      String regex = fixupRegexChar(delimit);      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = list.split(regex, -1);
      
      return array.length;
   }

   /**
    * Returns the number of comma-delimited entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The number of entries in the list.
    */
   public static int numEntriesOf(String list)
   {
      return numEntriesOf(list, ',');
   }

   /**
    * Returns the number of comma-delimited entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(character list)
   {
      if (list.isUnknown())
      {
         return new integer();
      }
      
      return new integer(numEntriesOf(list.getValue()));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(character list, String delimit)
   {
      if (list.isUnknown())
      {
         return new integer();
      }
      
      return new integer(numEntriesOf(list.getValue(), TextOps.safeDelimiter(delimit)));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(String list, character delimit)
   {
      if (delimit.isUnknown())
      {
         return new integer();
      }
      
      return new integer(numEntriesOf(list, TextOps.safeDelimiter(delimit.getValue())));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list or the
    *           <code>unknown value</code> if
    *           any of the inputs are <code>unknown value</code>.
    */
   public static integer numEntries(character list, character delimit)
   {
      if (list.isUnknown() || delimit.isUnknown())
      {
         return new integer();
      }
      
      return new integer(numEntriesOf(list.getValue(),
                                      TextOps.safeDelimiter(delimit.getValue())));
   }

   /**
    * Returns the number of entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    * @param    delimit
    *           The delimiter character.
    *
    * @return   The number of entries in the list.
    */
   public static integer numEntries(String list, char delimit)
   {
      return new integer(numEntriesOf(list, delimit));
   }

   /**
    * Returns the number of comma-delimited entries in the passed list.  
    *
    * @param    list
    *           The delimited list of entries.
    *
    * @return   The number of entries in the list.
    */
   public static integer numEntries(String list)
   {
      return new integer(numEntriesOf(list));
   }

   /**
    * Perform positional replacement of '&amp;n' specifiers in the format string
    * <code>fmt</code>, using the array of <code>character</code> and/or
    * <code>String</code> objects as the list of replacements.  The 1-based
    * index position of the '&amp;n' specifiers (where n and range from 1 to 
    * 9) is converted to a 0-based index into the <code>replacements</code>
    * array.  Instances of '&amp;&amp;' yield a single '&amp;' in the result.
    * Any specifiers that do not correspond to an object in the array are
    * replaced with an empty string. 
    *
    * @param    fmt
    *           The string specifying the text and format specifiers. 
    * @param    replacements
    *           The list of replacements.
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>fmt</code> string is <code>unknown value</code>.
    */
   public static character substitute(character fmt, Object[] replacements)
   throws ErrorConditionException
   {
      if (fmt.isUnknown())
         return new character();
      
      return substitute(fmt.getValue(), replacements);
   }

   /**
    * Perform positional replacement of '&amp;n' specifiers in the format string
    * <code>fmt</code>, using the array of <code>character</code> and/or
    * <code>String</code> objects as the list of replacements.  The 1-based
    * index position of the '&amp;n' specifiers (where n and range from 1 to 
    * 9) is converted to a 0-based index into the <code>replacements</code>
    * array.  Instances of '&amp;&amp;' yield a single '&amp;' in the result.
    * Any specifiers that do not correspond to an object in the array are
    * replaced with an empty string. 
    *
    * @param    fmt
    *           The string specifying the text and format specifiers. 
    * @param    replacements
    *           The list of replacements.
    *
    * @return   The resulting string.
    */
   public static character substitute(String fmt, Object[] replacements)
   throws ErrorConditionException
   {
      StringBuffer sb = new StringBuffer();
      
      int len  = fmt.length();
      
      for (int i = 0; i < len; i++)
      {
         char c = fmt.charAt(i);
         
         if (c == '&') 
         {
            // ignore this char and look at the next one
            i++;
            c = fmt.charAt(i);
            
            if (c >= '0' && c <= '9')
            {
               // convert this character into a 0-based index into the array
               int idx = Character.digit(c, 10) - 1;
               
               // check if this is a valid replacement index (note that 0 is
               // always ignored/invalid), if not the result is an empty
               // string since we don't make any replacement in that case
               if (idx >= 0 && idx < replacements.length) 
               {
                  Object obj = replacements[idx];
                  
                  if (obj != null)
                  {
                     // make a replacement if it is an object we recognize
                     if (obj instanceof String)
                     {
                        sb.append((String) obj);   
                     }
                     else if (obj instanceof character)
                     {
                        character ch = (character) obj;
                        sb.append(ch.isUnknown() ? "?" : ch.getValue());
                     }
                     else if (obj instanceof BaseDataType)
                     {
                        BaseDataType bdt = (BaseDataType) obj;
                        sb.append(bdt.toStringExport());
                     }
                  }
               }
            }            
            else if (c == '&')
            {
               sb.append(c);
            }
            else
            {
               String spec = "Invalid character \'%c\' following \'&\'" +
                             " in SUBSTITUTE";
               ErrorManager.recordOrThrowError(2920, String.format(spec, c));
               
               // this is the safest thing to return (it should be undone anyway)
               return new character("");
            }
         }
         else
         {
            // no match, put the char into the output
            sb.append(c);
         }
      }
      
      return new character(sb.toString());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     NumberType pos,
                                     NumberType len,
                                     String     type)
   {
      if (src.isUnknown() || pos.isUnknown() || len.isUnknown())
         return new character();
      
      return substring(src.getValue(),
                       pos.intValue(),
                       len.intValue(),
                       type,
                       src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     NumberType pos,
                                     NumberType len,
                                     character  type)
   {
      if (src.isUnknown() || pos.isUnknown() || len.isUnknown() || type.isUnknown())
         return new character();
      
      return substring(src.getValue(),
                       pos.intValue(),
                       len.intValue(),
                       type.getValue(),
                       src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     int        pos,
                                     NumberType len,
                                     String     type)
   {
      if (src.isUnknown() || len.isUnknown())
         return new character();
      
      return substring(src.getValue(),
                       pos,
                       len.intValue(),
                       type,
                       src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     int        pos,
                                     NumberType len,
                                     character  type)
   {
      if (src.isUnknown() || len.isUnknown() || type.isUnknown())
         return new character();
      
      return substring(src.getValue(),
                       pos,
                       len.intValue(),
                       type.getValue(),
                       src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     NumberType pos,
                                     int        len,
                                     String     type)
   {
      if (src.isUnknown() || pos.isUnknown())
         return new character();
      
      return substring(src.getValue(),
                       pos.intValue(),
                       len,
                       type,
                       src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     NumberType pos,
                                     int        len,
                                     character  type)
   {
      if (src.isUnknown() || pos.isUnknown() || type.isUnknown())
         return new character();
      
      return substring(src.getValue(),
                       pos.intValue(),
                       len,
                       type.getValue(),
                       src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if the
    *           <code>src</code> is <code>unknown</code>.
    */
   public static character substring(character src,
                                     int       pos,
                                     int       len,
                                     String    type)
   {
      if (src.isUnknown())
         return new character();
      
      return substring(src.getValue(), pos, len, type, src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring or <code>unknown value</code> if the
    *           <code>src</code> is <code>unknown</code>.
    */
   public static character substring(character src,
                                     int       pos,
                                     int       len,
                                     character type)
   {
      if (src.isUnknown() || type.isUnknown())
         return new character();
      
      return substring(src.getValue(), pos, len, type.getValue(), src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    * <p>
    * The returned instance is case-insensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    *
    * @return   The resulting substring.
    */
   public static character substring(String src,
                                     int    pos,
                                     int    len,
                                     String type)
   {
      return substring(src, pos, len, type, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     NumberType pos,
                                     NumberType len)
   {
      if (src.isUnknown() || pos.isUnknown() || len.isUnknown())
         return new character();
      
      return substring(src.getValue(),
                       pos.intValue(),
                       len.intValue(),
                       null,
                       src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     int        pos,
                                     NumberType len)
   {
      if (src.isUnknown() || len.isUnknown())
         return new character();
      
      return substring(src.getValue(), pos, len.intValue(), null, src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character  src,
                                     NumberType pos,
                                     int        len)
   {
      if (src.isUnknown() || pos.isUnknown())
         return new character();
      
      return substring(src.getValue(), pos.intValue(), len, null, src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    *
    * @return   The resulting substring or <code>unknown value</code> if the
    *           <code>src</code> is <code>unknown</code>.
    */
   public static character substring(character src, int pos, int len)
   {
      if (src.isUnknown())
         return new character();
      
      return substring(src.getValue(), pos, len, null, src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * extending for the rest of the string.   Operates on characters rather
    * than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(character src, NumberType pos)
   {
      if (src.isUnknown() || pos.isUnknown())
         return new character();
      
      return substring(src.getValue(), pos.intValue(), -1, null, src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * extending for the rest of the string.  Operates on characters rather
    * than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    *
    * @return   The resulting substring or <code>unknown value</code> if the
    *           <code>src</code> is <code>unknown</code>.
    */
   public static character substring(character src, int pos)
   {
      if (src.isUnknown())
         return new character();
      
      return substring(src.getValue(), pos, -1, null, src.isCaseSensitive());
   }

   /**
    * Returns the substring of the string at the given 1-based index  and
    * extending for the rest of the string. Operates on characters rather than
    * bytes.
    * <p>
    * The returned instance is case-insensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts.
    *
    * @return   The resulting substring.
    */
   public static character substring(String src, int pos)
   {
      return substring(src, pos, -1, null, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(String     src,
                                     NumberType pos,
                                     NumberType len)
   {
      if (pos.isUnknown() || len.isUnknown())
         return new character();
      
      return substring(src, pos.intValue(), len.intValue(), null, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(String     src,
                                     NumberType pos,
                                     int        len)
   {
      if (pos.isUnknown())
         return new character();
      
      return substring(src, pos.intValue(), len, null, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * Case-sensitivity of the returned instance is determined by the
    * <code>src</code> instance case-sensitivity.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string. 
    *
    * @return   The resulting substring or <code>unknown value</code> if any
    *           input is <code>unknown</code>.
    */
   public static character substring(String     src,
                                     int        pos,
                                     NumberType len)
   {
      if (len.isUnknown())
         return new character();
      
      return substring(src, pos, len.intValue(), null, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.  Operates on characters rather than bytes.
    * <p>
    * The returned instance is case-insensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    *
    * @return   The resulting substring.
    */
   public static character substring(String src, int pos, int len)
   {
      return substring(src, pos, len, null, false);
   }

   /**
    * Returns the substring of the string at the given 1-based index and
    * of the given length.
    *
    * @param    src
    *           The string to operate upon.
    * @param    pos
    *           The 1-based index at which the substring starts. 
    * @param    len
    *           The length of the substring to return. Use -1 for the rest of
    *           the string.
    * @param    type
    *           Specifies whether this substring operates on bytes or
    *           characters.  At this time only "raw" and "character" are
    *           supported.
    * @param    caseSens
    *           Determines the case-sensitivity of the returned instance.
    *
    * @return   The resulting substring.
    */
   public static character substring(String  src,
                                     int     pos, 
                                     int     len,
                                     String  type,
                                     boolean caseSens)
   {
      String result = null;
      
      // convert the Progress 1-based index into Java's 0-based index
      int start = pos - 1;
      
      if (pos < 1)
      {
         genNegativeIndexError();
         return new character("");
      }
      
      if (type != null && "raw".equalsIgnoreCase(type))
      {
         // raw mode (operates on bytes)
         byte[] buffer = src.getBytes(); 
         
         if (start >= buffer.length)
         {
            // requesting data past the end of the string yields the empty
            // string
            result = "";
         }
         else
         {
            int end = start + len;
            
            if (len == -1 || end >= buffer.length)
            {
               len = buffer.length - start;
            }
            
            result = new String(buffer, start, len);         
         }
      }
      else
      {      
         // character mode
         
         if (start >= src.length())
         {
            // requesting data past the end of the string yields the empty
            // string
            result = "";
         }
         else
         {
            int end = start + len;
            
            if (len == -1 || end >= src.length())
            {
               // if the requested data is more than the rest of the string
               // or if the rest of the string was explicitly requested via
               // -1
               result = src.substring(start);
            }
            else
            {
               // only return a portion
               result = src.substring(start, end);
            }            
         }         
      }
      
      return new character(result, caseSens);
   }

   /**
    * Returns a character instance that contains the lowercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    * If the operand is case-sensitive, the result will be case-sensitive.
    *
    * @param    c
    *           The character instance to convert (cannot be
    *           <code>null</code>).
    *
    * @return   The lowercased character instance (which should match up
    *           with the Progress equivalents in the ASCII range) or
    *           <code>unknown value</code> if the input string is
    *           <code>unknown value</code>.
    */
   public static character toLowerCase(character c)
   {
      if (c.isUnknown())
         return new character();
      
      character res = toLowerCase(c.getValue());
      
      // honor case-sensitivity
      res.setCaseSensitive(c.isCaseSensitive());
      
      return res;
   }

   /**
    * Returns a character instance that contains the lowercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    * The result will be case-insensitive.
    *
    * @param    str
    *           The string to convert (cannot be <code>null</code>).
    *
    * @return   The lowercased character instance which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   public static character toLowerCase(String str)
   {
      return new character(str.toLowerCase());
   }

   /**
    * Returns a character instance that contains the uppercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    * If the operand is case-sensitive, the result will be case-sensitive.
    *
    * @param    c
    *           The character instance to convert (cannot be
    *           <code>null</code>).
    *
    * @return   The uppercased character instance (which should match up
    *           with the Progress equivalents in the ASCII range) or
    *           <code>unknown value</code> if the input string is
    *           <code>unknown value</code>.
    */
   public static character toUpperCase(character c)
   {
      if (c.isUnknown())
         return new character();
      
      character res = toUpperCase(c.getValue());
      
      // honor case-sensitivity
      res.setCaseSensitive(c.isCaseSensitive());
      
      return res;
   }

   /**
    * Returns a character instance that contains the uppercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case 
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    *The result will be case-insensitive.
    *
    * @param    str
    *           The string to convert (cannot be <code>null</code>).
    *
    * @return   The uppercased character instance which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   public static character toUpperCase(String str)
   {
      return new character(str.toUpperCase());
   }

   /**
    * Returns a string instance that contains the uppercased version of
    * the specified string. This may generate incorrect results in a
    * DBCS environment since it will use UNICODE as the basis for case
    * conversion. If this returned value is different than would be returned
    * under Progress AND if that Progress-specific value is used in
    * comparison with this returned value, then a difference in logic will be
    * seen.
    * <p>
    * This method does not handle the case of user-specified source and/or
    * target codepages as is possible in Progress.
    * <p>
    *The result will be case-insensitive.
    *
    * @param    str
    *           The string to convert (cannot be <code>null</code>).
    * @param    runtimeString
    *           <code>true</code> if the target string is a runtime string,
    *           <code>false</code> if it is a hardcoded string into a source
    *           file.
    *
    * @return   The uppercased character instance which should match up
    *           with the Progress equivalents in the ASCII range.
    */
   public static String toUpperCase(String str, boolean runtimeString)
   {
      str = str.toUpperCase();
   
      if (!runtimeString)
      {
         StringBuffer sb = new StringBuffer("");
         int end = str.length();
   
         for (int i = 0; i < end; i++)
         {
            char current = str.charAt(i);
            char next;
   
            switch (current)
            {
               case '\\':
                  i++;
                  next = str.charAt(i);
                  sb.append('\\');
   
                  // change back mistakenly uppercased special characters
                  // \t \r \n \b \f into a hardcoded string
                  switch (next)
                  {
                     case 'T':
                        sb.append('t');
                        break;
                     case 'R':
                        sb.append('r');
                        break;
                     case 'N':
                        sb.append('n');
                        break;
                     case 'B':
                        sb.append('b');
                        break;
                     case 'F':
                        sb.append('f');
                        break;
                     default:
                        sb.append(next);
                  }
                  break;
               default:
                  sb.append(current);
            }
         }
   
         str = sb.toString();
      }
   
      return str;
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of only
    * the input parameter, the trim character list case-sensitivity is
    * ignored.  Likewise, if the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character trim(character input, character list)
   {
      if (input.isUnknown() || list.isUnknown())
      {
         return new character();
      }
      
      return trim(input.getValue(), list.getValue(), input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * input parameter. Likewise, if the input parameter is case-sensitive,
    * then the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character trim(character input, String list)
   {
      if (input.isUnknown())
      {
         return new character();
      }
      
      return trim(input.getValue(), list, input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string.
    * <p>
    * Trim operations are performed case-insensitively and the returned
    * result will be case-insensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character trim(String input, character list)
   {
      if (list.isUnknown())
      {
         return new character();
      }
      
      return trim(input, list.getValue(), false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * left and right of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * input parameter.  Likewise, if the input parameter is case-sensitive,
    * then the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character trim(character input)
   {
      if (input.isUnknown())
      {
         return new character();
      }
      
      return trim(input.getValue(), null, input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string. If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   public static character trim(String input, String list, boolean caseSens)
   {
      // trim left
      String result = trim(input, list, true, caseSens);
      
      // trim right
      return new character(trim(result, list, false, caseSens), caseSens);
   }

   /**
    * Trims the list of contiguous trim characters from the left and right
    * of an input string. If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    *
    * @return   The trimmed result.
    */
   public static character trim(String input, String list)
   {
      return trim(input, list, false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * left and right of an input string.  Trim operations are performed
    * case-insensitively.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.    
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character trim(String input)
   {
      return trim(input, (String) null, false);
   }

   /**
    * Trims the list of contiguous trim characters from the left or right of
    * an input string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    left
    *           Specifies if the trim should occur on the left or right
    *           side of the string.
    * @param    caseSens
    *           If <code>true</code>, a case-sensitive trim is performed.
    *
    * @return   The trimmed result.
    */
   static String trim(String  input,
                      String  list,
                      boolean left,
                      boolean caseSens)
   {
      // implement our default list
      if (list == null)
      {
         list = " \t\n\r";
      }
      
      int len  = input.length();
      int llen = list.length();
      int i;
      
      for (i = left ? 0 : len - 1; left ? i < len : i >= 0;)
      {
         boolean found = false;
         char    c     = input.charAt(i);
         char    cu    = Character.toUpperCase(c);
         
         for (int j = 0; j < llen; j++)
         {
            char ch  = list.charAt(j);
            char chu = Character.toUpperCase(ch);
            
            if ((caseSens && ch == c) || (!caseSens && chu == cu))  
            {
               found = true;
               break;
            }
         }
         
         if (!found)
         {
            break;
         }
         
         if (left)
         {
            i++;
         }
         else
         {
            i--;
         }
      }
      
      return left ? input.substring(i) : input.substring(0, i + 1);
   }

   /**
    * Given a delimiter string, extract the first character and retrun it.  If
    * the string is empty, return a space character.
    * 
    * @param   delimiter
    *          Delimiter string
    * 
    * @return  Delimiter character as described above.
    */
   static char safeDelimiter(String delimiter)
   {
      return (delimiter.length() == 0 ? ' ' : delimiter.charAt(0));
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.  
    * <p>
    * The search and the returned instance are case-sensitive if any of the
    * inputs are case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code>.
    */
   public static character replaceAll(character src,
                                      character from,
                                      character to)
   {
      if (src.isUnknown() || from.isUnknown() || to.isUnknown())
      {
         return new character();
      }
      
      boolean cs = (src.isCaseSensitive() || from.isCaseSensitive() || to.isCaseSensitive());
      
      return replaceAll(src.getValue(), from.getValue(), to.getValue(), cs);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.  
    * <p>
    * The search and the returned instance are case-sensitive if any of the
    * inputs are case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code>.
    */
   public static character replaceAll(character src,
                                      character from,
                                      String    to)
   {
      if (src.isUnknown() || from.isUnknown())
      {
         return new character();
      }
      
      boolean cs = (src.isCaseSensitive() || from.isCaseSensitive());
      
      return replaceAll(src.getValue(), from.getValue(), to, cs);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.  
    * <p>
    * The search and the returned instance are case-sensitive if any of the
    * inputs are case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code>.
    */
   public static character replaceAll(character src,
                                      String    from,
                                      character to)
   {
      if (src.isUnknown() || to.isUnknown())
      {
         return new character();
      }
      
      boolean cs = (src.isCaseSensitive() || to.isCaseSensitive());
      
      return replaceAll(src.getValue(), from, to.getValue(), cs);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.  
    * <p>
    * The search and the returned instance are case-sensitive if any of the
    * inputs are case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if any
    *           input is <code>unknown value</code>.
    */
   public static character replaceAll(String    src,
                                      character from,
                                      character to)
   {
      if (from.isUnknown() || to.isUnknown())
      {
         return new character();
      }
      
      boolean cs = (from.isCaseSensitive() || to.isCaseSensitive());
      
      return replaceAll(src, from.getValue(), to.getValue(), cs);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    * <p>
    * The search and the returned instance are case-sensitive if the
    * <code>src</code> instance is case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>src</code> string is <code>unknown value</code>.
    */
   public static character replaceAll(character src, String from, String to)
   {
      if (src.isUnknown())
      {
         return new character();
      }
      
      return replaceAll(src.getValue(), from, to, src.isCaseSensitive());
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    * <p>
    * The search and the returned instance are case-sensitive if the
    * <code>from</code> instance is case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>from</code> string is <code>unknown value</code>.
    */
   public static character replaceAll(String src, character from, String to)
   {
      if (from.isUnknown())
      {
         return new character();
      }
      
      return replaceAll(src, from.getValue(), to, from.isCaseSensitive());
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    * <p>
    * The search and the returned instance are case-sensitive if the
    * <code>to</code> instance is case-sensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>to</code> string is <code>unknown value</code>.
    */
   public static character replaceAll(String src, String from, character to)
   {
      if (to.isUnknown())
      {
         return new character();
      }
      
      return replaceAll(src, from, to.getValue(), to.isCaseSensitive());
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    * <p>
    * The search and the returned instance are always case-insensitive.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    *
    * @return   The resulting string or <code>unknown value</code> if the
    *           <code>src</code> string is <code>unknown value</code>.
    */
   public static character replaceAll(String src, String from, String to)
   {
      return replaceAll(src, from, to, false);
   }

   /**
    * Searches for all matches to the <code>from</code> string in the
    * <code>src</code> string and replaces these with the <code>to</code>
    * string.
    *
    * @param    src
    *           The string to operate upon.
    * @param    from
    *           The substring to search for. 
    * @param    to
    *           The substring to replace when a match is found. 
    * @param    cs
    *           Determines if a case-sensitive match should be performed AND
    *           the case-sensitivity of the returned instance.
    *
    * @return   The resulting string.
    */
   public static character replaceAll(String  src,
                                      String  from,
                                      String  to,
                                      boolean cs)
   {
      StringBuffer sb = new StringBuffer();
      
      int len  = src.length();
      int flen = from.length(); 
      
      for (int i = 0; i < len; i++)
      {
         if (src.regionMatches(!cs, i, from, 0, flen)) 
         {
            // make a replacement
            sb.append(to);
            
            // move past the end of the replaced text (take into account that
            // as the loop iterates, i will be incremented again)
            i += (flen - 1);
         }
         else
         {
            // no match, put the char into the output
            sb.append(src.charAt(i));
         }
      }
      
      return new character(sb.toString(), cs);
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of only
    * the input parameter, the trim character list case-sensitivity is
    * ignored.  Likewise, if the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(character input, character list)
   {
      if (input.isUnknown() || list.isUnknown())
      {
         return new character();
      }
      
      return rightTrim(input.getValue(), list.getValue(), input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * the input parameter. If the input parameter is case-sensitive, then
    * the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(character input, String list)
   {
      if (input.isUnknown())
      {
         return new character();
      }
      
      return rightTrim(input.getValue(), list, input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.
    * <p>
    * Trim operations are performed based case-insensitively.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(String input, character list)
   {
      if (list.isUnknown())
      {
         return new character();
      }
      
      return rightTrim(input, list.getValue(), false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * right of an input string.
    * <p>
    * Trim operations are performed based on the case-sensitivity of the
    * input parameter.  Likewise, if the input parameter is case-sensitive,
    * then the returned instance will be case-sensitive.
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(character input)
   {
      if (input.isUnknown())
      {
         return new character();
      }
      
      return rightTrim(input.getValue(), null, input.isCaseSensitive());
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    * @param    caseSens
    *           Specifies the case-sensitivity of the trim operation AND of
    *           the returned instance.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(String  input,
                                     String  list,
                                     boolean caseSens)
   {
      return new character(trim(input, list, false, caseSens), caseSens);
   }

   /**
    * Trims the list of contiguous trim characters from the right of an input
    * string.  If the list of trim characters is <code>null</code>
    * the method defaults to space, tab, line feed and carriage return.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.
   
    *
    * @param    input
    *           The string to trim.
    * @param    list
    *           The list of characters to trim or <code>null</code> to
    *           use the default list.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(String input, String list)
   {
      return rightTrim(input, list, false);
   }

   /**
    * Trims any contiguous space, tab, line feed and carriage return from the
    * right of an input string.
    * <p>
    * Trim operations and the returned instance will be case-insensitive.    
    *
    * @param    input
    *           The string to trim.
    *
    * @return   The trimmed result.
    */
   public static character rightTrim(String input)
   {
      return rightTrim(input, (String) null);
   }   
}
